# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import logging
import os
import re
import sys
import threading
import time

import colorama
import inotify_simple
import pexpect

import fingertip
import fingertip.util.log
from fingertip.plugins.backend.qemu import NotEnoughSpaceForSnapshotException


WATCHER_DEBOUNCE_TIMEOUT = .25  # seconds
CHECKPOINT_SPARSITY = 2  # seconds
TRAIL_EATING_TIMEOUT = .01  # seconds
MID_PHASE_TIMEOUT = .25  # seconds
DELAY_BEFORE_SEND = .01  # seconds


# Dimming colors #

def dim(text):
    text = text.replace(colorama.Style.NORMAL, colorama.Style.DIM)
    text = text.replace(colorama.Style.BRIGHT, colorama.Style.NORMAL)
    text = text.replace(colorama.Style.RESET_ALL,
                        colorama.Style.RESET_ALL + colorama.Style.DIM)
    return colorama.Style.DIM + text


# File monitoring and waiting #

class RewindNeededException(Exception):
    pass


def _is_event_rerun_worthy(event):
    if event.name.startswith('.#'):  # hi, EMACS
        return False
    if event.name.startswith('.') and event.name.endswith('.swp'):  # hi, VIM
        return False
    if event.name.startswith('.') and event.name.endswith('.swx'):  # why, VIM?
        return False
    return True


class OneOffInotifyWatcher:
    def __init__(self, log):
        self._inotify = inotify_simple.INotify()
        self.rewind_needed = threading.Event()
        self._mask = inotify_simple.flags.MODIFY
        self._mask |= inotify_simple.flags.ATTRIB
        self._mask |= inotify_simple.flags.CLOSE_WRITE
        self._mask |= inotify_simple.flags.MOVED_TO
        self._mask |= inotify_simple.flags.CREATE
        self._mask |= inotify_simple.flags.DELETE_SELF

        def _event_loop():
            log.debug(f'inotify blocks')
            for event in self._inotify.read():
                log.debug(f'inotify {event}')
                if not _is_event_rerun_worthy(event):
                    continue
                log.debug(f'that was rerun-worthy')

            # exhausting the events queue / debouncing
            debounce_end_time = time.time() + WATCHER_DEBOUNCE_TIMEOUT
            while True:
                time_left = debounce_end_time - time.time()
                if time_left < 0:
                    break
                _ = self._inotify.read(timeout=time_left)
            # finally, set the flag and cease functioning
            self.rewind_needed.set()
        threading.Thread(target=_event_loop, daemon=True).start()

    def watch(self, path):
        self._inotify.add_watch(path, mask=self._mask)


# some data classes #

class Segment:
    def __init__(self, input, expected_patterns):
        self.input, self.expected_patterns = input, expected_patterns

    def __eq__(self, other):
        return ((self.input, self.expected_patterns) ==
                (other.input, other.expected_patterns))


class SegmentExecutionResult:
    def __init__(self, segment, brief_output, full_output,
                 encountered_pattern, duration, checkpoint_after):
        self.input = segment.input
        self.expected_patterns = segment.expected_patterns
        self.encountered_pattern = encountered_pattern
        self.brief_output, self.full_output = brief_output, full_output
        self.duration = duration
        self.checkpoint_after = checkpoint_after  # can be None

    def corresponds_to(self, segment):
        return ((self.input, self.expected_patterns) ==
                (segment.input, segment.expected_patterns))


# adding operation on segments to a Machine #

def make_m_segment_aware(m):
    m.results = []
    m.checkpoint_sparsity = CHECKPOINT_SPARSITY
    m.never_executed_anything = True
    m.in_fast_forward = False

    def checkpoint_positions():  # which results have checkpoints after them
        return [i for i, res in enumerate(m.results) if res.checkpoint_after]
    m.checkpoint_positions = checkpoint_positions

    def since_last_checkpoint(results):
        duration = 0
        for res in results:
            duration = 0 if res.checkpoint_after else duration + res.duration
        return duration

    def checkpoint_cleanup():
        if not any((res.checkpoint_after is not None for res in m.results)):
            m.log.error('Error: ran out of checkpoints to clean up!')
            return
        deleted_checkpoints = 0
        m.checkpoint_sparsity *= 2
        m.log.debug(f'checkpoint cleanup, now with {m.checkpoint_sparsity} sec')
        for i in range(len(m.results)):
            if m.results[i].checkpoint_after:
                since_prev_checkpoint = (since_last_checkpoint(m.results[:i])
                                         + m.results[i].duration)
                if since_prev_checkpoint <= m.checkpoint_sparsity:
                    m.log.debug(f'deleting checkpoint after {i} ' +
                                f'({int(since_prev_checkpoint * 1000)}ms) '
                                'to free space')
                    deleted_checkpoints += 1
                    m.snapshot.remove(m.results[i].checkpoint_after)
                    m.results[i].checkpoint_after = None
                else:
                    m.log.debug(f' keeping checkpoint after {i} '
                                f'({int(since_prev_checkpoint * 1000)}ms)')
        if not deleted_checkpoints:
            checkpoint_cleanup()

    def maybe_checkpoint_already():
        latest_checkpoint_age = since_last_checkpoint(m.results)
        if not m.results or latest_checkpoint_age > m.checkpoint_sparsity:
            checkpoint_name = f'after-{len(m.results)}'
            try:
                m.log.debug(f'checkpointing after {checkpoint_name}')
                m.snapshot.checkpoint(checkpoint_name)
            except NotEnoughSpaceForSnapshotException:
                checkpoint_cleanup()  # try to remove at least one
                m.snapshot.checkpoint(checkpoint_name)
            return checkpoint_name

    def execute_segment(segment, checkpoint=True):
        start_time = time.time()

        m.log.debug(f'sending {repr(segment.input)}')
        m.never_executed_anything = False
        pre = ''
        if segment.input is not None:
            m.console.sendline(segment.input)
            # expect the input to appear (why not), but avoid line-wrapping
            m.console.expect_exact(segment.input.lstrip()[:40])
            ignored, pre = m.console.before, m.console.after
            if ignored.lstrip():
                m.log.warning(f'(ignored: {repr(ignored)})')
        else:
            m.console.sendcontrol('d')
        m.log.debug(f'sent {repr(segment.input)}')

        while True:
            try:
                i = m.console.expect(segment.expected_patterns,
                                     timeout=MID_PHASE_TIMEOUT)
                break
            except pexpect.exceptions.TIMEOUT:
                m.consider_interrupt_and_rewind()

        m.log.debug('-'*80)
        m.log.debug(segment.expected_patterns[i])
        m.log.debug('-'*80)

        end_time = time.time()
        result = SegmentExecutionResult(
            segment=segment,
            brief_output=pre + m.console.before,
            full_output=pre + m.console.before + m.console.after,
            encountered_pattern=segment.expected_patterns[i],
            duration=(end_time - start_time),
            checkpoint_after=None  # set later, see below
        )
        m.results.append(result)
        if checkpoint:
            result.checkpoint_after = maybe_checkpoint_already()
        m.consider_interrupt_and_rewind()
    m.execute_segment = execute_segment

    def eat_trailing():
        m.console.expect(pexpect.TIMEOUT, timeout=TRAIL_EATING_TIMEOUT)
        m.log.debug(f'ate trailing output: {repr(m.console.before)}')
    m.console.eat_trailing = eat_trailing

    def rewind_before_segment(i):
        if i == 0 and m.never_executed_anything:
            m.log.debug(f'clean VM, no rewind needed')
            return
        m.snapshot.freeze()
        for res in m.results[i:]:
            if res.checkpoint_after:
                m.log.debug(f'removing snapshot {res.checkpoint_after}...')
                m.snapshot.remove(res.checkpoint_after)
        m.log.debug(f'preparing for a rewind...')
        eat_trailing()
        m.console.buffer = ''
        m.log.info(f'rewinding before segment {i}')
        m.snapshot.revert(m.results[i-1].checkpoint_after
                          if i else m.snapshot.base_name)
        m.snapshot.unfreeze()
        m.results = m.results[:i]
    m.rewind_before_segment = rewind_before_segment

    def count_matching_segments(segments):
        for i, (segment, result) in enumerate(zip(segments, m.results)):
            if not result.corresponds_to(segment):
                return i
        return len(m.results)

    def reexecute(segments, watcher, reloader, terse, halt):
        def change_affects_current_segment():
            # we're in the process of executing a segment, did it change?
            updated_segments = reloader()
            matching_segments_n = count_matching_segments(updated_segments)
            if matching_segments_n < len(m.results):
                m.log.info('rewind due to changes in previous segments')
                return True
            elif matching_segments_n > len(segments):
                m.log.info('rare overexecution-after-truncation, rewinding')
                return True
            elif len(segments) > matching_segments_n:  # we have a current one
                assert len(m.results) == matching_segments_n < len(segments)
                currently_executing_segment = updated_segments[len(m.results)]
                replacement_segment = segments[len(m.results)]
                if currently_executing_segment != replacement_segment:
                    m.log.debug(f'rewind affects current segment')
                    return True
            return False

        def consider_interrupt_and_rewind():
            if watcher.rewind_needed.is_set():
                if change_affects_current_segment():
                    m.log.info('interrupt!')
                    raise RewindNeededException()
                # TODO: rate-limit rechecks?
        m.consider_interrupt_and_rewind = consider_interrupt_and_rewind

        i = count_matching_segments(segments)  # first mismatching segment idx
        if i == len(segments):
            m.log.debug(f'no changes, not reexecuting anything, {i}/{i}')
            return  # None -> no changes
        # i is now pointing at the first mismatching/missing segment,
        # but it doesn't necessarily mean we have a checkpoint right before it
        while i > 0 and not m.results[i-1].checkpoint_after:
            i -= 1  # 0th segment is guaranteed to have one, the base one
            m.log.debug(f'slipping back past uncheckpointed segment {i}')
        # i is now pointing at the closest segment with a checkpoint
        m.rewind_before_segment(i)

        # pseudo fast-forward
        m.in_fast_forward = True
        sys.stderr.flush()
        sys.stdout.flush()
        m.log.debug('---')
        if os.getenv('FINGERTIP_DEBUG') != '1':
            os.system('clear')
        previous_output = ''.join([
            result.full_output
            for result in m.results[:i]
        ])
        # TODO: use strip_control_sequences?
        m.console.logfile_read.write(dim(previous_output))
        m.in_fast_forward = False

        # execute the rest for real
        for j, segment in enumerate(segments[i:], i):
            if j == 0:
                m.console.logfile_read.write(m.repl_header)
                m.console.logfile_read.flush()
            last = j == len(segments) - 1
            m.log.debug(f'Executing segment {j} for real:')
            m.execute_segment(segment, checkpoint=not last)
            if j == 0:
                m.results[0].full_output = (m.repl_header +
                                            m.results[0].full_output)
            if any((halt(l) for l in m.results[-1].full_output.split('\n'))):
                m.log.warning('halting (use --halt-on to configure that, '
                              'e.g., --halt-on=nothing)')
                return True
        return True  # -> Completed till the end, executing some changes
    m.reexecute = reexecute


# Different REPLs support #


class REPLBase:
    RETCODE_MARKER = '\u200Creturn code '
    RETCODE_MATCH = r'\u200Creturn code \d+'

    @classmethod
    def segment(cls, code):
        lines = code.split('\n')
        segments = [Segment(s, [cls.rPS1, cls.rPS2]) for s in lines]
        segments.append(Segment(None, [cls.RETCODE_MATCH + '\r+\n']))
        return segments

    @classmethod
    def install_interpreter_if_missing(cls, m):
        if m(f'command -v {cls.INTERPRETER}', check=False).retcode:
            m.apply('ansible', 'package',
                    name=['bash', cls.PACKAGE], state='installed')
        m(f'command -v {cls.INTERPRETER}')

    @classmethod
    def launch_interpreter(cls, m, scriptpath, cmd=None, pre=''):
        m.console.sendline(f'PS1=""')
        m.console.sendline(f"bash -c '{pre} exec -a {scriptpath} "
                           f"{cmd or cls.INTERPRETER}'; "
                           r'bash -c "echo -e ' + cls.RETCODE_MARKER + '$?"')

    @classmethod
    def _filter_generic(cls, line, terseness):
        if terseness:
            if line in (cls.PS1, cls.PS2):
                return False
            if line.startswith(cls.PS1 + cls.COMMENT_SIGN):
                return False
            if line.startswith(cls.PS2 + cls.COMMENT_SIGN):
                return False
        if terseness in ('more', 'most'):
            if line.startswith(cls.PS2):
                return False
        if terseness == 'most':
            if line.startswith(cls.PS1) or line.startswith(cls.PS2):
                return False
            if re.match(cls.RETCODE_MATCH, line):
                return False
        return True

    @classmethod
    def format(cls, line):
        if cls.line_is_an_input(line):
            return colorama.Fore.BLUE + line
        elif line == cls.RETCODE_MARKER + '0':
            return colorama.Fore.GREEN + line
        elif re.match(cls.RETCODE_MATCH, line):
            return colorama.Fore.MAGENTA + line
        elif cls.line_is_an_error(line):
            return colorama.Fore.RED + line
        elif cls.line_is_a_warning(line):
            return colorama.Fore.YELLOW + line
        return line

    @staticmethod
    def line_is_a_warning(l):
        return False

    @staticmethod
    def line_is_an_error(l):
        return False


class REPLBash(REPLBase):
    INTERPRETER = PACKAGE = 'bash'
    INTERPRETER_ARGS = '--noprofile --norc'
    PS1, PS2 = '\u200C$ ', '\u200C> '
    rPS1, rPS2 = r'\u200c\$ ', r'\u200c> '
    COMMENT_SIGN = '#'

    @classmethod
    def prepare(cls, m, scriptpath, terse):
        with m:
            cls.install_interpreter_if_missing(m)

            if terse != 'most':
                bash_version = m('bash --version').out.split('\n')[0]
                m.repl_header = bash_version + '\n' + cls.PS1
            else:
                m.repl_header = cls.PS1

            # trick taken from pexpect.replwrap
            # this is not visible in actual PS1, but visible in, e.g., env
            TRICK = '\\[\\]'
            cls.launch_interpreter(m, scriptpath, 'bash --noprofile --norc',
                                   pre=(f'PS1="{TRICK}{cls.PS1}" '
                                        f'PS2="{TRICK}{cls.PS2}"'))
            m.console.sendline(r'echo -e \\u200C""READY')
            m.console.expect(cls.rPS1 + r'echo -e \\\\u200C""READY\r+\n'
                             r'\u200cREADY\r+\n' + cls.rPS1)
        return m

    @classmethod
    def line_is_an_input(cls, line):
        return line.startswith(cls.PS1) or line.startswith(cls.PS2)

    @classmethod
    def filter(cls, line, terseness):
        return cls._filter_generic(line, terseness)


class REPLPython(REPLBase):
    INTERPRETER = PACKAGE = 'python3'
    INTERPRETER_ARGS = ''
    PS1, PS2 = '\u200C>>> ', '\u200C... '
    rPS1, rPS2 = r'\u200C>>> ', r'\u200C... '
    COMMENT_SIGN = '#'

    @classmethod
    def segment(cls, code):
        if code.split('\n')[-1][:1].isspace():  # indented last line?
            code += '\n'  # terminate an open '...' to execute it
        return super().segment(code)

    @classmethod
    def prepare(cls, m, scriptpath, terse):
        with m:
            cls.install_interpreter_if_missing(m)
            cls.launch_interpreter(m, scriptpath)

            m.console.expect(r'\r+\n(Python.*?)\r\n')
            if terse != 'most':
                m.repl_header = m.console.match.group(1) + '\n' + cls.PS1
            else:
                m.repl_header = cls.PS1

            m.console.sendline('import sys')
            m.console.sendline(f'sys.ps1, sys.ps2 = "{cls.PS1}", "{cls.PS2}"')
            m.console.sendline('del sys')
            m.console.sendline(r'print("\u200C" + "READY")')
            m.console.expect(r'\r+\n\u200CREADY\r+\n' + cls.rPS1)
        return m

    @staticmethod
    def line_is_an_error(line):
        return bool(re.match(r'\w*Error: ', line))

    @staticmethod
    def line_is_a_warning(line):
        return bool(re.search(r':\d+:.*Warning: ', line))

    @classmethod
    def line_is_an_input(cls, line):
        return line.startswith(cls.PS1) or line.startswith(cls.PS2)

    @classmethod
    def format(cls, line):
        if line == 'Traceback (most recent call last):':  # not an error yet
            return colorama.Fore.RED + line
        elif re.match(r'  File ".*", line \d+, in ', line):  # not an error yet
            return colorama.Fore.RED + line
        return super().format(line)

    @classmethod
    def filter(cls, line, terseness):
        if terseness in ('more', 'most'):
            if line == 'Traceback (most recent call last):':
                return False
        return cls._filter_generic(line, terseness)


repls = {
    'bash': REPLBash,
    'python': REPLPython,
}


@fingertip.transient
def main(m, scriptpath, language='bash', unseal=True,
         terse=False, color=True, script_reading_hook=None,
         halt_on='warning'):
    script_reading_hook = script_reading_hook or (lambda m, code: code)
    m = m.apply('unseal') if unseal else m
    # m = m.apply('.hooks.disable_proxy')

    repl = repls[language] if isinstance(language, str) else language
    repl = repl()
    m = m.apply(repl.prepare, scriptpath, terse)

    def reloader():
        with open(scriptpath) as f:
            code = script_reading_hook(m, f.read())
        return repl.segment(code)

    if halt_on == 'warning':
        def halt(line):
            return repl.line_is_an_error(line) or repl.line_is_a_warning(line)
    elif halt_on == 'error':
        halt = repl.line_is_an_error
    else:
        def halt(line):
            return False

    fingertip.util.log.plain()

    if os.getenv('FINGERTIP_DEBUG') != '1':
        if color and hasattr(repl, 'format'):
            class Formatter(logging.Formatter):
                def format(self, record):
                    formatted = repl.format(record.msg)
                    if hasattr(m, 'in_fast_forward') and m.in_fast_forward:
                        formatted = dim(formatted)
                    return formatted + colorama.Style.RESET_ALL
            fingertip.util.log.current_handler.setFormatter(Formatter())
        if terse and hasattr(repl, 'filter'):
            class Filter(logging.Filter):
                def filter(self, record):
                    return repl.filter(record.msg, terse)
            fingertip.util.log.current_handler.addFilter(Filter())

    with m:
        make_m_segment_aware(m)

        # FIXME: changes aren't reuploaded!
        m.ssh.upload(scriptpath)  # make self-referential code a bit happier

        # disable input echoing
        m.console.logfile_read = m.console.logfile
        m.console.logfile = None
        # speed up
        m.console.delaybeforesend = DELAY_BEFORE_SEND

        while True:
            try:
                watcher = OneOffInotifyWatcher(m.log)
                watcher.watch(scriptpath)
                segments = reloader()
                any_changes = m.reexecute(segments, watcher, reloader, terse,
                                          halt)
                if any_changes:
                    m.console.eat_trailing()
                    if not terse:
                        m.log.info('(too much output? try --terse,'
                                   ' --terse=more or --terse=most)')
                    checkpoints_i = [str(i) for i in m.checkpoint_positions()]
                    m.log.info(f'done. checkpoints: {checkpoints_i}, '
                               f'sparsity={m.checkpoint_sparsity}s. '
                               'waiting for changes...')
                watcher.rewind_needed.wait()
            except RewindNeededException:
                continue
            except KeyboardInterrupt:
                break



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions and constants for fingertip: reflinking (CoW-powered copying).
"""
# TODO: clean up even more?

import os
import sys
import subprocess

from fingertip.util import log, temp, path


SETUP = os.getenv('FINGERTIP_SETUP', 'suggest')
SIZE = os.getenv('FINGERTIP_SETUP_SIZE', '25G')


def _cp_reflink(src, dst, *args):
    subprocess.run(['cp', '-rT', *args, src, dst], check=True)


def always(src, dst, preserve=False):
    args = ['--preserve=all'] if preserve else []
    _cp_reflink(src, dst, '--reflink=always', *args)


def auto(src, dst, preserve=False):
    args = ['--preserve=all'] if preserve else []
    _cp_reflink(src, dst, '--reflink=auto', *args)


def is_supported(dirpath):
    tmp = temp.disappearing_file(dstdir=dirpath, create=True)
    r = subprocess.Popen(['cp', '--reflink=always', tmp, tmp + '-reflink'],
                         stderr=subprocess.PIPE)
    _, err = r.communicate()
    r.wait()
    temp.remove(tmp, tmp + '-reflink')
    sure_not = b'failed to clone' in err and b'Operation not supported' in err
    if r.returncode and not sure_not:
        log.error('reflink support detection inconclusive, cache dir problems')
        log.error(f'({err})')
    return r.returncode == 0


def create_supported_fs(backing_file, size):
    subprocess.run(['fallocate', '-l', size, backing_file], check=True)
    subprocess.run(['mkfs.xfs', '-m', 'reflink=1', backing_file],
                   check=True)


def mount_supported_fs(backing_file, tgt):
    log.info('mounting a reflink-supported filesystem for image storage...')
    tgt_uid, tgt_gid = os.stat(tgt).st_uid, os.stat(tgt).st_gid
    subprocess.run(['sudo', 'mount', '-o', 'loop', backing_file, tgt],
                   check=True)
    mount_uid, mount_gid = os.stat(tgt).st_uid, os.stat(tgt).st_gid
    if (tgt_uid, tgt_gid) != (mount_uid, mount_gid):
        log.debug(f'fixing owner:group ({tgt_uid}:{tgt_gid})')
        subprocess.run(['sudo', 'chown', f'{tgt_uid}:{tgt_gid}', tgt],
                       check=True)
        if tgt.startswith('/home'):
            subprocess.run(['sudo', 'semanage', 'fcontext', '-a', '-t',
                            'user_home_dir_t', tgt + '(/.*)?'], check=False)
            subprocess.run(['sudo', 'restorecon', '-v', tgt], check=False)


def storage_setup_wizard():
    assert SETUP in ('auto', 'suggest', 'never')
    if SETUP == 'never':
        return
    size = SIZE
    os.makedirs(path.MACHINES, exist_ok=True)
    if not is_supported(path.MACHINES):
        log.warning(f'images directory {path.MACHINES} lacks reflink support')
        log.warning('without it, fingertip will thrash and fill up your SSD '
                    'in no time')
        if not os.path.exists(path.COW_IMAGE):
            if SETUP == 'suggest':
                log.info(f'would you like to allow fingertip '
                         f'to allocate {size} at {path.COW_IMAGE} '
                         'for a reflink-enabled XFS loop mount?')
                log.info('(set FINGERTIP_SETUP="auto" environment variable'
                         ' to do it automatically)')
                i = input(f'[{size}]/different size/cancel/ignore> ').strip()
                if i == 'cancel':
                    log.error('cancelled')
                    sys.exit(1)
                elif i == 'ignore':
                    return
                size = i or size
            tmp = temp.disappearing_file(os.path.dirname(path.COW_IMAGE))
            create_supported_fs(tmp, size)
            os.rename(tmp, path.COW_IMAGE)

        log.info(f'fingertip will now mount the XFS image at {path.COW_IMAGE}')
        if SETUP == 'suggest':
            i = input(f'[ok]/skip/cancel> ').strip()
            if i == 'skip':
                log.warning('skipping; '
                            'fingertip will have no reflink superpowers')
                log.warning('tell your SSD I\'m sorry')
                return
            elif i and i != 'ok':
                log.error('cancelled')
                sys.exit(1)

        mount_supported_fs(path.COW_IMAGE, path.CACHE)


def storage_unmount():
    log.plain()
    log.info(f'unmounting {path.CACHE} ...')
    subprocess.run(['sudo', 'umount', '-l', path.CACHE])
    log.nicer()


def storage_destroy():
    mount = subprocess.run(['mount'], capture_output=True)
    if path.COW_IMAGE in mount.stdout.decode():
        log.warning('Filesystem is still mounted. Trying to unmount.')
        storage_unmount()

    if os.path.exists(path.COW_IMAGE):
        os.unlink(path.COW_IMAGE)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


@fingertip.transient
def main(m):
    with m:
        r = m('hostname -f')
        assert r.out.endswith('.fingertip.local\n')
        assert m('hostname -d').out == 'fingertip.local\n'



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import os
import shutil
import tempfile
import time

import fasteners

import fingertip.expiration
import fingertip.machine
from fingertip.util import log, path, temp, units


def _time(path):
    s = os.stat(path)
    return max(s.st_mtime, s.st_atime, s.st_ctime)


@fingertip.transient
def main(what=None, *args, **kwargs):
    if what == 'everything':
        return everything()
    if what == 'periodic':
        return periodic()
    elif what in ('downloads', 'logs', 'machines', 'tempfiles'):
        return globals()[what](*args, **kwargs)
    log.error('usage: ')
    log.error('    fingertip cleanup downloads [<older-than>]')
    log.error('    fingertip cleanup logs [<older-than>]')
    log.error('    fingertip cleanup machines [<expired-for>|all]')
    log.error('    fingertip cleanup tempfiles [<older-than> [<location]]')
    log.error('    fingertip cleanup everything')
    log.error('    fingertip cleanup periodic')
    raise SystemExit()


def downloads(older_than=0):
    cutoff_time = time.time() - units.parse_time_interval(older_than)
    _cleanup_dir(path.DOWNLOADS, lambda f: _time(f) >= cutoff_time)


def logs(older_than=0):
    cutoff_time = time.time() - units.parse_time_interval(older_than)
    _cleanup_dir(path.LOGS, lambda f: _time(f) >= cutoff_time)


def _cleanup_dir(dirpath, preserve_func):
    for root, dirs, files in os.walk(dirpath, topdown=False):
        for f in (os.path.join(root, x) for x in files):
            assert os.path.realpath(f).startswith(os.path.realpath(dirpath))
            if not preserve_func(f):
                log.info(f'removing {os.path.realpath(f)}')
                os.unlink(f)
        for d in (os.path.join(root, x) for x in dirs):
            assert os.path.realpath(d).startswith(os.path.realpath(dirpath))
            try:
                os.rmdir(d)
            except OSError:  # directory not empty => ignore
                pass


def machines(expired_for=0):
    if expired_for != 'all':
        adjusted_time = time.time() - units.parse_time_interval(expired_for)
    for root, dirs, files in os.walk(path.MACHINES, topdown=False):
        for d in (os.path.join(root, x) for x in dirs):
            lock_path = os.path.join(root, '.' + os.path.basename(d) + '-lock')
            lock = fasteners.process_lock.InterProcessLock(lock_path)
            lock.acquire()
            try:
                remove = fingertip.machine.needs_a_rebuild(d, by=adjusted_time)
            except Exception as ex:
                log.warning(f'while processing {d}: {ex}')
                remove = True
            if (expired_for == 'all' or remove):
                assert os.path.realpath(d).startswith(
                    os.path.realpath(path.MACHINES)
                )
                log.info(f'removing {os.path.realpath(d)}')
                if not os.path.islink(d):
                    shutil.rmtree(d)
                else:
                    os.unlink(d)
            else:
                log.debug(f'keeping {os.path.realpath(d)}')
            os.unlink(lock_path)
            lock.release()


def tempfiles(older_than='6h', location=None):
    location = location or tempfile.gettempdir()
    cutoff_time = time.time() - units.parse_time_interval(older_than)
    _cleanup_dir(location, lambda f: (_time(f) >= cutoff_time or
                                      temp.AUTOREMOVE_PREFIX not in f))


def periodic():
    machines('6h')
    downloads('30d')
    logs('30d')
    tempfiles()
    tempfiles(location='/tmp')  # backend.qemu uses /tmp specifically


def everything():
    downloads()
    logs()
    machines('all')
    tempfiles(0)
    tempfiles(0, '/tmp')  # backend.qemu uses /tmp specifically



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import datetime
import os
import sys
import time

from fingertip.util import log, weak_hash, units


class Expiration:
    def __init__(self, expire_in):
        self.time = time.time() + units.parse_time_interval(expire_in)
        self._deps_files, self._deps_dirs = {}, {}

    def pretty(self):
        return datetime.datetime.fromtimestamp(self.time).isoformat()

    def cap(self, interval):
        self.time = min(self.time,
                        time.time() + units.parse_time_interval(interval))

    def is_expired(self, by=None):
        return self.time < (by or time.time())

    def depend_on(self, tree):
        if not os.path.isdir(tree):
            self.depend_on_a_file(tree)
        else:
            for root, dirs, files in os.walk(tree):
                self.depend_on_a_directory_listing(root)
                for d in dirs:
                    self.depend_on_a_directory_listing(os.path.join(root, d))
                for f in files:
                    self.depend_on_a_file(os.path.join(root, f))

    def depend_on_a_directory_listing(self, path):
        hash_ = weak_hash.of_string(repr(sorted(os.listdir(path))))
        self._deps_dirs[path] = (os.stat(path).st_mtime, hash_)

    def depend_on_a_file(self, path):
        path = os.path.abspath(path)
        if ((path.startswith('/usr/lib') or '/site-packages/' in path)
                and '/fingertip/' not in path):
            return
        self._deps_files[path] = (os.stat(path).st_mtime,
                                  weak_hash.of_file(path))

    def file_has_not_changed(self, path):
        log.debug(f'checking that {path} has not changed...')
        mtime, hash_ = self._deps_files[path]
        try:
            if mtime != (os.stat(path).st_mtime):
                if hash_ != weak_hash.of_file(path):
                    log.warning(f'{path} has changed, set '
                                'FINGERTIP_IGNORE_FILE_CHANGES=1 to ignore')
                    return False
        except FileNotFoundError:
            return False
        return True

    def dir_listing_has_not_changed(self, path):
        log.debug(f'checking that {path} listing has not changed...')
        mtime, hash_ = self._deps_dirs[path]
        try:
            if mtime != (os.stat(path).st_mtime):
                listing = repr(sorted(os.listdir(path)))
                if hash_ != weak_hash.of_string(listing):
                    log.warning(f'{path} listing has changed, set '
                                'FINGERTIP_IGNORE_FILE_CHANGES=1 to ignore')
                    return False
        except FileNotFoundError:
            return False
        return True

    def files_have_not_changed(self):
        if (os.getenv('FINGERTIP_IGNORE_CODE_CHANGES', '0') != '0' or
                os.getenv('FINGERTIP_IGNORE_FILE_CHANGES', '0') != '0'):
            return True
        return (all((self.file_has_not_changed(p) for p in self._deps_files))
                and
                all((self.dir_listing_has_not_changed(p)
                    for p in self._deps_dirs)))

    def depend_on_loaded_python_modules(self):
        """
        Make the machine depend on all python modules loaded to date
        in an imprecise best-effort overly-cautious attempt
        to get automatic rebuilds triggered on source code changes.
        """
        for module in sys.modules:
            try:
                module_file = sys.modules[module].__file__
                if module_file:
                    self.depend_on_a_file(module_file)
            except AttributeError:
                pass



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

import logging
import os
import re
import subprocess

import fingertip
from fingertip.util import reflink


META_TEMPLATE = '''
instance-id: fingertip
local-hostname: {HOSTNAME}
'''

USER_TEMPLATE = '''
#cloud-config
fqdn: {FQDN}
final_message: cloud-config final message
ssh_authorized_keys:
  - {SSH_PUBKEY}
disable_root: False
chpasswd:
  list: |
     root:fingertip
  expire: False
'''


def main(m=None, url=None):
    m = m or fingertip.build('backend.qemu')
    assert url
    assert hasattr(m, 'qemu')

    # because we have no idea how to unseal it later
    m.sealed = False
    m.expiration.cap('4h')

    image_file = os.path.join(m.path, os.path.basename(url))
    if '://' in url:
        m.log.info(f'fetching {url}...')
        m.http_cache.fetch(url, image_file)
    else:
        m.log.info(f'copying {url}...')
        reflink.auto(url, image_file)
        m.expiration.depend_on_a_file(url)

    m.log.info('resizing image...')
    run = m.log.pipe_powered(subprocess.run,
                             stdout=logging.INFO, stderr=logging.ERROR)
    run(['qemu-img', 'resize', image_file, m.qemu.disk_size], check=True)
    m.qemu._image_to_clone = image_file
    m.qemu.virtio_scsi = True  # in case it's Linux <5

    with m:
        hostname = url.rsplit('/', 1)[-1].rsplit('.', 1)[0].replace('.', '_')
        hostname = hostname.replace('x86_64', '')
        fqdn = hostname + '.fingertip.local'

        meta_data = META_TEMPLATE.format(FQDN=fqdn,
                                         HOSTNAME=hostname,
                                         SSH_PUBKEY=m.ssh.pubkey)
        meta_file = os.path.join(m.path, 'meta-data')
        with open(meta_file, 'w') as f:
            f.write(meta_data)
        m.http_cache.serve_local_file('/cloud-init/meta-data', meta_file)

        user_data = USER_TEMPLATE.format(FQDN=fqdn,
                                         HOSTNAME=hostname,
                                         SSH_PUBKEY=m.ssh.pubkey)
        user_file = os.path.join(m.path, 'user-data')
        with open(user_file, 'w') as f:
            f.write(user_data)
        m.http_cache.serve_local_file('/cloud-init/user-data', user_file)

        init_url = m.http_cache.internal_url + '/cloud-init/'
        seed = ['-smbios', f'type=1,serial=ds=nocloud-net;s={init_url}']

        m.qemu.run(load=None, extra_args=seed)

        m.console.expect_exact('cloud-config final message')
        m.console.sendline('')
        m.console.sendline('')
        m.console.expect(f'login:')
        m.console.sendline('root')
        m.console.expect('Password: ')
        m.console.sendline('fingertip')
        m.console.sendline(' echo prompt" "detection\n')
        m.console.expect_exact('prompt detection')
        m.prompt = re.search(r'\n(.+?) echo prompt', m.console.before).group(1)
        m.log.debug(f'm.prompt = {repr(m.prompt)}')
        m.console.sendline('')
        m.console.expect_exact(m.prompt)

        m.ram.safeguard = '512M'  # sane for 2020, and it's overrideable anyway

        def login(username='root', password='fingertip'):
            m.console.expect(f'login: ')
            m.console.sendline(username)
            m.console.expect('Password: ')
            m.console.sendline(password)
            m.console.expect_exact(m.prompt)
        m.login = login

        m.hooks.timesync.append(lambda: m('hwclock -s'))

        m.log.info('cloud-init finished')
    return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

from fingertip.machine import build, transient

__all__ = ['build', 'transient']



"""
Helper functions for fingertip: repeatedly keep trying.
"""

import time


def keep_trying(func, exception_types, retries=12, timeout=1/32):
    while retries:
        try:
            return func()
        except exception_types:
            retries -= 1
            if not retries:
                raise
            time.sleep(timeout)
            timeout *= 2



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import git

import os
import shutil
import tarfile


from fingertip.util import log, lock, path, reflink, temp


SAVIOUR_DEFAULTS = 'local,cached+direct'


def saviour_sources():
    return [(t[len('cached+'):] if t.startswith('cached+') else t,
             t.startswith('cached+'))
            for t in
            os.getenv('FINGERTIP_SAVIOUR', SAVIOUR_DEFAULTS).split(',')]


def _remove(p):
    assert p.startswith(path.DOWNLOADS)
    if os.path.exists(p):
        shutil.rmtree(p)


class Repo(git.Repo, lock.Lock):
    def __init__(self, url, *path_components, enough_to_have=None):
        assert path_components
        self.url = url
        cache_path = path.downloads('git', *path_components, makedirs=True)
        self.path = temp.disappearing_dir(os.path.dirname(cache_path),
                                          path_components[-1])
        lock_working_copy_path = self.path + '-lock'
        lock_cache_path = cache_path + '-lock'
        lock.Lock.__init__(self, lock_working_copy_path)
        update_not_needed = None
        sources = saviour_sources()
        self.self_destruct = False
        with lock.Lock(lock_cache_path), lock.Lock(lock_working_copy_path):
            cache_is_enough = False
            if os.path.exists(cache_path):
                try:
                    cr = git.Repo(cache_path)
                    cache_is_enough = enough_to_have and (
                        enough_to_have in (t.name for t in cr.tags) or
                        enough_to_have in (h.name for h in cr.heads) or
                        enough_to_have in (c.hexsha for c in cr.iter_commits())
                        # that's not all revspecs, but best-effort is fine
                    )
                except git.GitError as e:
                    log.error(f'something wrong with git cache {cache_path}')
                    log.error(str(e))
                _remove(self.path)

            for i, (source, cache) in enumerate(sources):
                last_source = i == len(sources) - 1

                if cache and cache_is_enough:
                    log.info(f'not re-fetching {url} from {source} '
                             f'because {enough_to_have} '
                             'is already present in cache')
                    git.Repo.clone_from(cache_path, self.path, mirror=True)
                    break

                if source == 'local':
                    surl = path.saviour(url).replace('//', '/')  # workaround
                    if not os.path.exists(surl) and not last_source:
                        continue
                    log.info(f'cloning {url} from local saviour mirror')
                    git.Repo.clone_from(surl, self.path, mirror=True)
                    break
                elif source == 'direct':
                    surl = url
                else:
                    surl = source + '/' + url
                    surl = 'http://' + surl if '://' not in source else surl

                log.info(f'cloning {url} from {source} '
                         f'cache_exists={os.path.exists(cache_path)}...')
                try:
                    # TODO: bare clone
                    # no harm in referencing cache, even w/o cached+
                    git.Repo.clone_from(surl, self.path, mirror=True,
                                        dissociate=True,
                                        reference_if_able=cache_path)
                except git.GitError:
                    log.warning(f'could not clone {url} from {source}')
                    if last_source:
                        raise
                    continue
                break

            _remove(cache_path)
            reflink.auto(self.path, cache_path)
            git.Repo.__init__(self, self.path)
            self.remotes[0].set_url(url)
        self.self_destruct = True

    def __enter__(self):
        lock.Lock.__enter__(self)
        git.Repo.__enter__(self)
        return self

    def __exit__(self, *args):
        if self.self_destruct:
            _remove(self.path)
        git.Repo.__exit__(self, *args)
        lock.Lock.__exit__(self, *args)


class Checkout(git.Repo, lock.Lock):
    def __init__(self, url, *path_components, enough_to_have=None):
        with Repo(url, *path_components, enough_to_have=enough_to_have) as r:
            cache_path = path.downloads('git', *path_components, makedirs=True)
            self.path = temp.disappearing_dir(os.path.dirname(cache_path),
                                              path_components[-1])
            self.self_destruct = False
            git.Repo.clone_from(r.path, self.path)
        git.Repo.__init__(self, self.path)

    def __enter__(self):
        git.Repo.__enter__(self)
        return self

    def __exit__(self, *args):
        git.Repo.__exit__(self, *args)
        _remove(self.path)


def upload_clone(m, url, path_in_m, rev=None, rev_is_enough=True):
    assert hasattr(m, 'ssh')
    with m:
        kwa = {} if not rev_is_enough else {'enough_to_have': rev}
        with Repo(url, url.replace('/', '::'), **kwa) as repo:
            tar = temp.disappearing_file()
            tar_in_m = f'/.tmp-{os.path.basename(tar)}'
            extracted_in_m = f'/.tmp-{os.path.basename(tar)}-extracted'
            log.info(f'packing {url} checkout...')
            with tarfile.open(tar, 'w') as tf:
                tf.add(repo.path, arcname=extracted_in_m)
            log.info(f'uploading {url} checkout...')
            m.ssh.upload(tar, tar_in_m)
        log.info(f'performing {url} checkout...')
        m(f'''
            set -uex
            tar xmf {tar_in_m} -C /
            mkdir -p {path_in_m}
            git clone -n {extracted_in_m} {path_in_m}
            cd {path_in_m}
            git remote set-url origin {url}
            git checkout {f'{rev}' if rev else 'origin/HEAD'}
            rm -rf {extracted_in_m}
            rm -f {tar_in_m}
        ''')
    return m


def upload_contents(m, url, path_in_m, rev=None, rev_is_enough=True):
    assert hasattr(m, 'ssh')
    with m:
        kwa = {} if not rev_is_enough else {'enough_to_have': rev}
        with Repo(url, url.replace('/', '::'), **kwa) as repo:
            tar = temp.disappearing_file()
            log.info(f'packing {url} contents at rev {rev}...')
            tar_in_m = f'/.tmp-{os.path.basename(tar)}'
            with open(tar, 'wb') as tf:
                repo.archive(tf, treeish=rev, prefix=path_in_m + '/')
            log.info(f'uploading {url} contents at rev {rev}...')
            m.ssh.upload(tar, tar_in_m)
        log.info(f'unpacking {url} contents at rev {rev}...')
        m(f'''
            set -uex
            tar xmf {tar_in_m} -C /
            rm -f {tar_in_m}
        ''')
    return m



"""
Helper functions for fingertip: attaching and calling hooks.
"""

import collections


class HookSet(list):
    def __call__(self, *args, **kwargs):
        return [hook(*args, **kwargs) for hook in self]

    def in_reverse(self, *args, **kwargs):
        return [hook(*args, **kwargs) for hook in self[::-1]]


class _HookManager(collections.defaultdict):
    def __getattr__(self, hook_type):
        return self[hook_type]


def HookManager():
    return _HookManager(HookSet)  # pickling shenanigans workaround


# hooks = HookManager()

# hooks['smth'].append(func)
# hooks.smth.append(func)

# hooks['smth']()
# hooks.smth()

# hooks['smth'].in_reverse()
# hooks.smth.in_reverse()



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


def setup(m, greeting='Hello!'):
    with m:
        m.console.sendline('PS1=@"parent> "')
        m.console.expect_exact('@parent> ')
        m.console.sendline('PS1=@"child> " sh')
        m.console.expect_exact('@child> ')
        return m


@fingertip.transient
def main(m, greeting='Hello!'):
    with m.apply(setup, greeting=greeting) as m:
        m.console.sendline('')
        m.console.expect_exact('@child> ')
        m.console.sendcontrol('d')
        m.console.expect_exact('@parent> ')



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import requests

import os

import fingertip.machine
from fingertip.util import http_cache, log, path


FEDORA_GEOREDIRECTOR = 'http://download.fedoraproject.org/pub/fedora/linux'
RELEASED = 33


def main(m=None, version=RELEASED, mirror=None, specific_mirror=True,
         fips=False):
    m = m or fingertip.build('backend.qemu')
    if hasattr(m, 'qemu'):
        m = m.apply(install_in_qemu, version=version, mirror=mirror,
                    specific_mirror=specific_mirror, fips=fips)
    elif hasattr(m, 'container'):
        m = m.apply(m.container.from_image, f'fedora:{version}')
        with m:
            m('dnf -y update')
    else:
        raise NotImplementedError()
    return m


def determine_mirror(mirror, version, releases_development):
    # if you have a saviour mirror, let's assume it's a good one
    for source, _ in http_cache.saviour_sources():
        if source != 'direct' and http_cache.is_fetcheable(source, mirror):
            return mirror
    # we can query a georedirector for a local Fedora mirror and use just
    # that one, consistently. problem is, it also yields really broken ones.
    # let's check that a mirror has at least a repomd.xml and a kernel:
    updates_repomd = f'updates/{version}/Everything/x86_64/repodata/repomd.xml'
    kernel = (f'{releases_development}/{version}'
              '/Everything/x86_64/os/isolinux/vmlinuz')

    h = requests.head(mirror + '/' + updates_repomd, allow_redirects=False)
    if h.status_code in (301, 302, 303, 307, 308) and 'Location' in h.headers:
        r = h.headers['Location'].rstrip('/').replace('https://', 'http://')
        assert r.endswith('/' + updates_repomd)
        base = r[:-len('/' + updates_repomd)]
        # good, now now ensure it also has a kernel
        h = requests.head(base + '/' + kernel)
        if h.status_code != 200:
            log.warning(f'{base + "/" + kernel} -> {h.status_code}')
            log.warning(f'mirror {base} is broken, trying another one')
            return determine_mirror(mirror, version, releases_development)
        else:
            return base
    return mirror


def install_in_qemu(m, version, mirror=None, specific_mirror=True, fips=False):
    version = int(version) if version != 'rawhide' else 'rawhide'
    releases_development = ('development'
                            if version in (RELEASED + 1, 'rawhide')
                            else 'releases')
    if mirror is None:
        if not specific_mirror:
            mirror = FEDORA_GEOREDIRECTOR  # not consistent, not recommended!
        else:
            mirror = determine_mirror(FEDORA_GEOREDIRECTOR, version,
                                      releases_development)
            m.log.info(f'autoselected mirror {mirror}')
    url = f'{mirror}/{releases_development}/{version}/Everything/x86_64/os'
    upd = f'{mirror}/updates/{version}/Everything/x86_64'
    repos = (f'url --url {url}\n' +
             f'repo --name fedora --baseurl {url}\n' +
             f'repo --name updates --baseurl {upd}')

    with m:
        m.ram.safeguard = '768M'
        m.expiration.cap('2d')  # non-immutable repositories

        hostname = f'fedora{version}' + ('-fips' if fips else '')
        fqdn = hostname + '.fingertip.local'
        ssh_key_fname = path.fingertip('ssh_key', 'fingertip.pub')
        with open(ssh_key_fname) as f:
            ssh_pubkey = f.read().strip()
        m.expiration.depend_on_a_file(ssh_key_fname)

        ks_fname = path.fingertip('kickstart_templates',
                                  f'fedora{version}' if version != 'rawhide'
                                  else f'fedora{RELEASED+1}')
        with open(ks_fname) as f:
            ks_text = f.read().format(HOSTNAME=fqdn,
                                      SSH_PUBKEY=ssh_pubkey,
                                      PROXY=m.http_cache.internal_url,
                                      MIRROR=mirror,
                                      REPOS=repos)
        m.expiration.depend_on_a_file(ks_fname)

        m.http_cache.mock('http://mock/ks', text=ks_text)
        m.log.info(f'fetching kernel: {url}/isolinux/vmlinuz')
        kernel = os.path.join(m.path, 'kernel')
        m.http_cache.fetch(f'{url}/isolinux/vmlinuz', kernel)
        m.log.info(f'fetching initrd: {url}/isolinux/initrd.img')
        initrd = os.path.join(m.path, 'initrd')
        m.http_cache.fetch(f'{url}/isolinux/initrd.img', initrd)
        append = ('ks=http://mock/ks inst.ks=http://mock/ks '
                  'inst.ksstrict '
                  'console=ttyS0 inst.notmux '
                  'inst.zram=off '
                  f'proxy={m.http_cache.internal_url} '
                  f'inst.proxy={m.http_cache.internal_url} '
                  f'inst.repo={url} ' +
                  ('fips=1' if fips else ''))
        extra_args = ['-kernel', kernel, '-initrd', initrd, '-append', append]

        with m.ram('3G'):
            m.qemu.run(load=None, extra_args=extra_args)
            i = m.console.expect(['Storing configuration files and kickstarts',
                                  'installation failed',
                                  'installation was stopped',
                                  'installer will now terminate'])
            assert i == 0, 'Installation failed'
            m.qemu.wait()

        m.qemu.run(load=None)  # cold boot

        def login(username='root', password='fingertip'):
            if username == 'root':
                m.prompt = f'[root@{hostname} ~]# '
            else:
                m.prompt = f'[{username}@{hostname} ~]$ '
            m.console.expect(f'{hostname} login: ')
            m.console.sendline(username)
            m.console.expect('Password: ')
            m.console.sendline(password)
            m.console.expect_exact(m.prompt)

        m.login = login

        m.login()
        m.log.info('Fedora installation finished')

        def disable_proxy():
            return m.apply('ansible', 'ini_file', path='/etc/dnf/dnf.conf',
                           section='main', option='proxy', state='absent')
        m.hooks.disable_proxy.append(disable_proxy)

        m.hooks.unseal += [lambda: m('systemctl restart NetworkManager'),
                           lambda: m('nm-online')]

        m.hooks.timesync.append(lambda: m('hwclock -s'))

        m.fedora = version
        m.dist_git_branch = f'f{version}' if version != 'rawhide' else 'master'

        return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions for fingertip: scheduling a cleanup job with systemd
"""
# TODO: clean up even more?

import shutil
import subprocess

from fingertip.util import log


def schedule():
    # Do this only if fingertip is in PATH
    if not shutil.which("fingertip"):
        log.debug('No `fingertip` found in PATH. Not scheduling '
                  'automatic cleanup.')
        return

    # Skip if systemd is not available
    if not shutil.which('systemd-run') or not shutil.which('systemctl'):
        log.warning('It looks like systemd is not available. '
                    'No cleanup is scheduled! If you are running out of disk, '
                    'space, run `fingertip cleanup periodic` manually.')
        return

    # If the timer is already installed skip installation too
    p = subprocess.run(['systemctl', '--user', 'is-active', '--quiet',
                       'fingertip-cleanup.timer'])
    if p.returncode == 0:
        log.debug('The systemd timer handling cleanup is already installed '
                  'and running.')
        return

    # Run twice a day
    log.info('Scheduling cleanup to run every two hours')
    subprocess.run(['systemd-run', '--unit=fingertip-cleanup', '--user',
                    '--on-calendar=0/2:00:00',
                    'fingertip', 'cleanup', 'periodic'])



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions and constants for fingertip and fingertip plugins: paths.
"""
# TODO: fold into caching_proxy, there should be very few users of this module

import contextlib
import os

import xdg.BaseDirectory

import fingertip.util.lock as util_lock


FINGERTIP = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
CACHE = os.path.join(xdg.BaseDirectory.xdg_cache_home, 'fingertip')
COW_IMAGE = os.path.join(CACHE, 'cow.xfs.img')
DOWNLOADS = os.path.join(CACHE, 'downloads')
SAVIOUR = os.path.join(CACHE, 'saviour')
MACHINES = os.path.join(CACHE, 'machines')
SHARED = os.path.join(CACHE, 'shared')
LOGS = os.path.join(CACHE, 'logs')


def easy_accessor(root_path):
    def easy_access_func(*more_path_components, makedirs=False):
        if makedirs:
            os.makedirs(os.path.join(root_path, *more_path_components[:-1]),
                        exist_ok=True)
        return os.path.join(root_path, *more_path_components)
    return easy_access_func


fingertip = easy_accessor(FINGERTIP)
cache = easy_accessor(CACHE)
downloads = easy_accessor(DOWNLOADS)
machines = easy_accessor(MACHINES)
saviour = easy_accessor(SAVIOUR)
shared = easy_accessor(SHARED)
logs = easy_accessor(LOGS)


@contextlib.contextmanager
def wip(normal_path, makedirs=False):
    norm, wip, lock = normal_path, normal_path + '-WIP', normal_path + '-lock'
    with util_lock.Lock(lock):
        if os.path.exists(norm):
            assert not os.path.exists(wip), f'Both {norm} and {wip} exist'
            os.rename(norm, wip)
        if makedirs:
            os.makedirs(os.path.dirname(wip.rstrip(os.path.sep)),
                        exist_ok=True)
        yield wip
        os.rename(wip, norm)
        os.unlink(lock)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import atexit
import contextlib
import json
import logging
import os
import selectors
import socket
import stat
import subprocess
import threading
import time
import queue

import fasteners
import pexpect

import fingertip.machine
import fingertip.util.http_cache
from fingertip.util import free_port, log, path, reflink, repeatedly, temp
from fingertip.util import units


IGNORED_EVENTS = ('NIC_RX_FILTER_CHANGED', 'RTC_CHANGE', 'RESET')
SNAPSHOT_BASE_NAME = 'tip'  # it has to have some name
CACHE_INTERNAL_IP, CACHE_INTERNAL_PORT = '10.0.2.244', 8080
CACHE_INTERNAL_URL = f'http://{CACHE_INTERNAL_IP}:{CACHE_INTERNAL_PORT}'
# TODO: add a way to customize smp
QEMU_COMMON_ARGS = ['-enable-kvm', '-cpu', 'host', '-smp', '4',
                    '-virtfs', f'local,id=shared9p,path={path.SHARED},'
                               'security_model=mapped-file,mount_tag=shared',
                    '-nographic',
                    '-object', 'rng-random,id=rng0,filename=/dev/urandom',
                    '-device', 'virtio-balloon',
                    '-device', 'virtio-rng-pci,rng=rng0']


def main(arch='x86_64', ram_min='1G', ram_size='1G', ram_max='4G',
         disk_size='20G', custom_args=[], guest_forwards=[]):
    assert arch == 'x86_64'
    m = fingertip.machine.Machine('qemu')
    m.arch = arch
    m._host_forwards = []
    m.ram = RAMNamespacedFeatures(m, ram_min, ram_size, ram_max)
    m.qemu = QEMUNamespacedFeatures(m, disk_size, custom_args)
    m.snapshot = SnapshotNamespacedFeatures(m)
    m._backend_mode = 'pexpect'

    # TODO: extract SSH into a separate plugin?
    m.ssh = SSH(m)

    def load():
        m.http_cache = fingertip.util.http_cache.HTTPCache()
        m.http_cache.internal_ip = CACHE_INTERNAL_IP
        m.http_cache.internal_port = CACHE_INTERNAL_PORT
        m.http_cache.internal_url = CACHE_INTERNAL_URL

        if hasattr(m.ram, '_pre_down_size'):
            m.ram._target = m.ram._pre_down_size
            del m.ram._pre_down_size
    m.hooks.load.append(load)

    def up():
        if m.qemu.live:
            m.qemu.run()
            m.time_desync.fix_if_needed()
    m.hooks.up.append(up)

    def down():
        if m.qemu.live:
            if m.ram.min != m.ram.actual:
                m.log.info('driving RAM down to '
                           f'({units.binary(m.ram.min)})')
                m.ram._pre_down_size = m.ram._target
                m.ram.size = m.ram.min  # blocking
            m.log.debug(f'save_live {m.qemu.monitor}, {m.qemu.monitor._sock}')
            m.qemu.monitor.pause()
            m.qemu.monitor.checkpoint()
            m.qemu.monitor.commit()
            m.qemu.monitor.quit()
            m.qemu._go_down()
    m.hooks.down.append(down)

    def drop():
        if m.qemu.live:
            m.log.debug(f'drop {m.qemu.monitor}, {m.qemu.monitor._sock}')
            m.qemu.monitor.quit()
            m.qemu._go_down()
        if m._transient and m.qemu.image and os.path.exists(m.qemu.image):
            os.unlink(m.qemu.image)
    m.hooks.drop.append(drop)

    def save():
        m.http_cache = None
    m.hooks.save.append(save)

    def clone(to_path):
        m.qemu._image_to_clone = os.path.join(m.path, 'image.qcow2')
    m.hooks.clone.append(clone)

    def disrupt():
        if m.qemu.live:
            m.ssh.invalidate()
    m.hooks.disrupt.append(disrupt)

    def forward_host_port(hostport, guestport):
        if m.qemu.live:
            m.qemu.monitor.forward_host_port(hostport, guestport)
        m._host_forwards.append((hostport, guestport))
    m.forward_host_port = forward_host_port

    m.breakpoint = lambda: m.apply('ssh', unseal=False)

    load()
    run = m.log.pipe_powered(subprocess.run,
                             stdout=logging.INFO, stderr=logging.ERROR)
    run(['qemu-img', 'create', '-f', 'qcow2', '-o', 'lazy_refcounts=on',
         os.path.join(m.path, 'image.qcow2'), disk_size], check=True)
    return m


class QEMUNamespacedFeatures:
    def __init__(self, vm, disk_size, custom_args):
        self.vm = vm
        self.live = False
        self.disk_size = disk_size
        self.custom_args = custom_args
        self.image, self._image_to_clone = None, None
        self.virtio_scsi = False  # flip before OS install for TRIM on Linux<5
        self._qemu = f'qemu-system-{self.vm.arch}'

    def run(self, load=SNAPSHOT_BASE_NAME, guest_forwards=[], extra_args=[]):
        if load:
            self.vm.time_desync.report(self.vm.time_desync.LARGE)
        run_args = ['-loadvm', load] if load else []

        self.monitor = Monitor(self.vm)
        run_args += ['-qmp', (f'tcp:127.0.0.1:{self.monitor.port},'
                              'server,nowait,nodelay')]

        self.vm.ssh.port = free_port.find()
        self.vm.shared_directory = SharedDirectory(self.vm)
        self.vm.exec = self.vm.ssh.exec
        host_forwards = [(self.vm.ssh.port, 22)] + self.vm._host_forwards
        host_forwards = [f'hostfwd=tcp:127.0.0.1:{h}-:{g}'
                         for h, g in host_forwards]
        cache_guest_forward = (CACHE_INTERNAL_IP, CACHE_INTERNAL_PORT,
                               f'nc 127.0.0.1 {self.vm.http_cache.port}')
        guest_forwards = guest_forwards + [cache_guest_forward]
        run_args += ['-device', 'virtio-net,netdev=net0', '-netdev',
                     ','.join(['user', 'id=net0'] + host_forwards +
                              (['restrict=yes'] if self.vm.sealed else []) +
                              [f'guestfwd=tcp:{ip}:{port}-cmd:{cmd}'
                               for ip, port, cmd in guest_forwards])]

        self.image = os.path.join(self.vm.path, 'image.qcow2')
        if self._image_to_clone:
            # let's try to use /tmp (which is, hopefully, tmpfs) for transients
            # if it looks empty enough
            cloned_to_tmp = False
            required_space = os.path.getsize(self._image_to_clone) + 2 * 2**30
            if self.vm._transient:
                # Would be ideal to have it global (and multiuser-ok)
                tmp_free_lock = path.cache('.tmp-free-space-check-lock')
                with fasteners.process_lock.InterProcessLock(tmp_free_lock):
                    if temp.has_space(required_space, where='/tmp'):
                        self.image = temp.disappearing_file(
                            '/tmp', hint='fingertip-qemu'
                        )
                        self.vm.log.info('preloading image to /tmp...')
                        reflink.auto(self._image_to_clone, self.image)
                        self.vm.log.info('preloading image to /tmp completed')
                        cloned_to_tmp = True
            if not cloned_to_tmp:
                reflink.auto(self._image_to_clone, self.image)
            self._image_to_clone = None
        if self.virtio_scsi:
            run_args += ['-device', 'virtio-scsi-pci',
                         '-device', 'scsi-hd,drive=hd',
                         '-drive', f'file={self.image},cache=unsafe,'
                                   'if=none,id=hd,discard=unmap']
        else:
            run_args += ['-drive', f'file={self.image},cache=unsafe,'
                                   'if=virtio,discard=unmap']

        run_args += ['-m', str(self.vm.ram.max // 2**20)]

        os.makedirs(path.SHARED, exist_ok=True)

        args = QEMU_COMMON_ARGS + self.custom_args + run_args + extra_args
        self.vm.log.debug(' '.join(args))
        if self.vm._backend_mode == 'pexpect':
            # start connecting/negotiating QMP, later starts auto-ballooning
            threading.Thread(target=self.monitor.connect, daemon=True).start()
            pexp = self.vm.log.pseudofile_powered(pexpect.spawn,
                                                  logfile=logging.INFO)
            self.vm.console = pexp(self._qemu, args, echo=False,
                                   timeout=None,
                                   encoding='utf-8', codec_errors='ignore')
            self.live = True
        elif self.vm._backend_mode == 'direct':
            subprocess.run([self._qemu, '-serial', 'mon:stdio'] + args,
                           check=True)
            # FIXME: autoballooning won't start w/o the monitor connection!
            self.live = False
            self._go_down()

    def wait(self):
        self.vm.console.expect(pexpect.EOF)
        self.vm.console.close()
        assert self.vm.console.exitstatus == 0, 'QEMU terminated with an error'
        self._go_down()
        self.live = False

    def _go_down(self):
        if self.live:
            self.vm.console = None
            self.vm.ssh.invalidate()
        del self.vm.exec
        del self.monitor


class SnapshotNamespacedFeatures:
    def __init__(self, vm):
        self.vm = vm
        self.base_name = SNAPSHOT_BASE_NAME
        self.list = [self.base_name]  # FIXME: not always true =/
        self._frozen = False

    def checkpoint(self, name=SNAPSHOT_BASE_NAME):
        self.vm.qemu.monitor.checkpoint(name)
        self.list.append(name)

    def freeze(self):
        self.vm.qemu.monitor.pause()
        self._frozen = True

    def revert(self, name=SNAPSHOT_BASE_NAME):
        self.vm.qemu.monitor.restore(name)
        if not self._frozen:
            self.vm.time_desync.fix_if_needed()

    def unfreeze(self):
        self.vm.qemu.monitor.resume()
        self._frozen = False
        self.vm.time_desync.fix_if_needed()

    def remove(self, name):
        self.vm.qemu.monitor.del_checkpoint(name)
        self.list.remove(name)

    def purge(self, keep_base=True):
        for name in self.list:
            if not (keep_base and name == self.base_name):
                self.vm.qemu.monitor.del_checkpoint(name)
        self.list = [self.base_name] if keep_base else []


class VMException(RuntimeError):
    pass


class UnknownVMException(VMException):
    pass


class NotEnoughSpaceForSnapshotException(VMException):
    pass


class Monitor:
    def __init__(self, vm, port=None):
        self.vm = vm
        self.port = port or free_port.find()
        self.vm.log.debug(f'monitor port {self.port}')
        self._sock = None
        self._queue = queue.Queue()
        self._ram_actual_changed = threading.Event()
        self._ram_target_changed = threading.Event()
        self._command_execution_lock = threading.RLock()

    def connect(self, retries=12, timeout=1/32):
        if self._sock is None:
            with self._command_execution_lock:
                self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                self.vm.log.debug(f'QMP connecting...')
                repeatedly.keep_trying(
                    lambda: self._sock.connect(('127.0.0.1', self.port)),
                    ConnectionRefusedError, retries=retries, timeout=timeout
                )
                self._execute('qmp_capabilities')
                threading.Thread(target=self._recv_thread, daemon=True).start()
                server_greeting = self._expect()
                assert set(server_greeting.keys()) == {'QMP'}
                self._expect({'return': {}})  # from qmp_capabilities
                self.vm.log.debug(f'QMP is ready')
            threading.Thread(target=self._ballooning_thread,
                             daemon=True).start()

    def _recv_thread(self):
        while self._sock:
            r = self._recv()
            if not r:
                break

            if 'event' in r:
                if r['event'] in IGNORED_EVENTS:
                    self.vm.log.debug(f'ignoring: QMP message {r}')
                    continue
                elif r['event'] == 'BALLOON_CHANGE':
                    assert 'data' in r and 'actual' in r['data']
                    self.vm.ram._actual = r['data']['actual']
                    self.vm.log.debug('memory ballooned to '
                                      f'{units.binary(self.vm.ram._actual)}')
                    self._ram_actual_changed.set()
                    continue

            self._queue.put(r)

    def _ballooning_thread(self):
        while self._sock:
            if self.vm.ram._actual != self.vm.ram._target:
                self.balloon(self.vm.ram._target)
                time.sleep(2)
            else:
                self._ram_target_changed.wait()
                self._ram_target_changed.clear()

    def _expect(self, what=None):  # None = nothing exact, caller inspects it
        reply = self._queue.get()
        self.vm.log.debug(f'expected: {what}')
        self.vm.log.debug(f'received: {reply}')
        if what is not None:
            assert reply == what
        return reply

    def _disconnect(self):
        with self._command_execution_lock:
            if self._sock:
                self._sock.close()
                self._sock = None
                self.vm.log.debug('qemu monitor disconnected')

    def _send(self, dictionary):
        self._sock.send(json.dumps(dictionary).encode())

    def _recv(self):
        r = b''
        try:
            while self._sock:
                r += self._sock.recv(1)
                if not r:
                    self._disconnect()
                    return
                if r[-1] == ord('\n'):
                    return json.loads(r.decode())
        except OSError:
            self._disconnect()

    def _execute(self, cmd, retries=12, timeout=1/32, **kwargs):
        self.connect(retries, timeout)
        self.vm.log.debug(f'executing: {cmd} {kwargs}')
        if not kwargs:
            self._send({'execute': cmd})
        else:
            self._send({'execute': cmd, 'arguments': kwargs})

    def _execute_human_command(self, cmd):
        with self._command_execution_lock:
            self._execute(f'human-monitor-command', **{'command-line': cmd})
            r = self._expect(None)
            while 'timestamp' in r and 'event' in r:
                r = self._expect(None)  # ignore these for now
            assert set(r.keys()) == {'return'}
            assert isinstance(r['return'], str)
            r = r['return']
            if 'Error while writing VM state: No space left on device' in r:
                raise NotEnoughSpaceForSnapshotException(r)
            elif r:
                self.vm.log.error(r)
                raise UnknownVMException(r)
            return r

    def resume(self):
        self.vm.time_desync.report(self.vm.time_desync.SMALL)
        with self._command_execution_lock:
            self._execute('cont')
            r = self._expect(None)
            assert set(r.keys()) == {'timestamp', 'event'}
            assert r['event'] == 'RESUME'
            self._expect({'return': {}})

    def pause(self):
        self.vm.hooks.disrupt()
        self.vm.time_desync.report(self.vm.time_desync.SMALL)
        with self._command_execution_lock:
            self._execute('stop')
            r = self._expect(None)
            assert set(r.keys()) == {'timestamp', 'event'}
            assert r['event'] == 'STOP'
            self._expect({'return': {}})

    def quit(self):
        self.vm.hooks.disrupt()
        if self._command_execution_lock.acquire(blocking=False):
            self._execute('quit')
            self._expect({'return': {}})
            self._disconnect()

    def checkpoint(self, name=SNAPSHOT_BASE_NAME):
        self._execute_human_command(f'savevm {name}')

    def restore(self, name=SNAPSHOT_BASE_NAME):
        self.vm.time_desync.report(self.vm.time_desync.SMALL
                                   if name != SNAPSHOT_BASE_NAME else
                                   self.vm.time_desync.LARGE)
        self.vm.hooks.disrupt()
        self._execute_human_command(f'loadvm {name}')

    def del_checkpoint(self, name):
        self._execute_human_command(f'delvm {name}')

    def commit(self):
        # saves the changes to current file only,
        # as live QEMU images are not CoW-backed with qcow2 mechanism,
        # just deduplicated with FS-level reflinks that QEMU is unaware of
        self._execute_human_command('commit all')

    def forward_host_port(self, hostport, guestport):
        self._execute_human_command('hostfwd_add '
                                    f'tcp:127.0.0.1:{hostport}-:{guestport}')

    def balloon(self, target_size):
        target_size = units.parse_binary(target_size)
        if self.vm.ram._actual != target_size:
            with self._command_execution_lock:
                self.vm.log.debug('sending a request to balloon to '
                                  f'{units.binary(target_size)} from '
                                  f'{units.binary(self.vm.ram._actual)}')
                self._execute(f'balloon', **{'value': target_size})
                self._expect({'return': {}})
        else:
            self.vm.log.debug('no ballooning needed, '
                              f'already at {units.binary(target_size)}')


class RAMNamespacedFeatures:
    def __init__(self, m, min_, target, max_):
        self._m = m
        self._min = units.parse_binary(min_)
        self._size = units.parse_binary(target)
        self._target = self._size
        self._max = units.parse_binary(max_)
        self._actual = self._max
        self._safeguard = 0

    def set_size_async(self, new):
        new = units.parse_binary(new)
        if new < self._min:
            self._m.log.warning(f'cannot set ram.size to {units.binary(new)}, '
                                'clipping to ram.min '
                                f'({units.binary(self._min)})')
            new = self._min
        if new < self._safeguard:
            self._m.log.warning(f"won't set ram.size to {units.binary(new)}, "
                                'clipping to ram.safeguard '
                                f'({units.binary(self._safeguard)})')
            new = self._safeguard
        if new > self._max:
            self._m.log.warning(f'cannot set ram.size to {units.binary(new)}, '
                                'clipping to ram.max '
                                f'({units.binary(self._max)})')
            new = self._max
        self._target = new
        if self._m.qemu.live and self._actual != new:
            self._m.log.debug(f'going to balloon to {units.binary(new)} '
                              f'from {units.binary(self._m.ram._actual)}')
            self._m.qemu.monitor._ram_target_changed.set()

    def wait_for_ballooning(self):
        if self._actual != self._target:
            start_time = time.time()
            while self._actual != self._target:
                self._m.qemu.monitor._ram_actual_changed.wait()
                self._m.qemu.monitor._ram_actual_changed.clear()
            ballooning_duration = time.time() - start_time
            if ballooning_duration > 2.5:
                self._m.log.warning(f'ballooning took {ballooning_duration}s, '
                                    'increasing ram.min may help with delays')
            else:
                self._m.log.debug(f'ballooning took {ballooning_duration}s')
                self._m.qemu.monitor._ram_target_changed.set()
        else:
            self._m.log.debug(f'ballooning not needed')

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, new):
        new = units.parse_binary(new)
        if new < self._safeguard:
            self._m.log.warning(f"won't set ram.min to {units.binary(new)}, "
                                'clipping to ram.safeguard '
                                f'({units.binary(self._safeguard)})')
            new = self._safeguard
        if new > self._max:
            self._m.log.warning(f'cannot set ram.min to {units.binary(new)}, '
                                'clipping to ram.max '
                                f'({units.binary(self._max)})')
            new = self._max
        self._m.log.debug(f'setting ram.min to {units.binary(new)}')
        self._min = new
        if self._target < new:
            self._m.log.debug(f'bumping ram.size to {units.binary(new)} '
                              'along with ram.min')
            self.size = new

    @property
    def size(self):
        return self._target

    @size.setter
    def size(self, new):
        self.set_size_async(new)
        if self._m.qemu.live:
            self.wait_for_ballooning()  # sync if up, async best-effort if down

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, new):
        self._m.log.error('cannot change ram.max dynamically, '
                          'use `backend.qemu --ram-max=100500M + ...`')
        return

    @property
    def safeguard(self):
        return self._safeguard

    @safeguard.setter
    def safeguard(self, new):
        self._safeguard = units.parse_binary(new)

    @property
    def actual(self):
        return self._actual

    @contextlib.contextmanager  # with m, m.ram('+2G'): ...
    def __call__(self, size, wait=True, wait_post=True):
        if isinstance(size, str) and size and size[0] in '+-':
            new_size = self.size + units.parse_binary(size)
        elif isinstance(size, str) and size and size.startswith('>='):
            new_size = max(self.size, units.parse_binary(size[2:]))
        else:
            new_size = units.parse_binary(size)

        old_size = self.size
        if wait:
            self.size = new_size
        else:
            self.set_size_async(new_size)
        yield
        if wait_post:
            self.size = old_size
        else:
            self.set_size_async(old_size)


class SharedDirectory:
    def __init__(self, m):
        self.m = m
        self.mount_count = 0

    def __enter__(self):
        if not self.mount_count:
            self.m('mkdir -p /shared && '
                   'mount -t 9p -o trans=virtio shared /shared '
                   '  -oversion=9p2000.L')
        self.mount_count += 1

    def __exit__(self, *_):
        self.mount_count -= 1
        if not self.mount_count:
            self.m('umount /shared || ! mountpoint /shared')


class SSH:
    _key_file = path.fingertip('ssh_key', 'fingertip')
    key_file_paramiko = path.fingertip('ssh_key', 'fingertip.paramiko')
    pubkey_file = path.fingertip('ssh_key', 'fingertip.pub')

    def __init__(self, m, host='127.0.0.1', port=None):
        self.host, self.port = host, port
        self.m = m
        self.m.log.debug(f'ssh port {self.port}')
        self._transport = None

    def connect(self, force_reconnect=False, retries=12, timeout=1/32):
        atexit.register(self.invalidate)
        import paramiko  # ... in parallel with VM spin-up
        if not force_reconnect and self._transport is not None:
            self._transport.send_ignore()
            if self._transport.is_authenticated():
                return  # the transport is already OK
        self._transport = None
        self.m.log.debug('waiting for the VM to spin up and offer SSH...')
        pubkey = paramiko.ECDSAKey.from_private_key_file(SSH.key_file_paramiko)

        def connect():
            self.m.log.debug('Trying to connect ...')
            t = paramiko.Transport((self.host, self.port))
            t.start_client()
            return t

        transport = repeatedly.keep_trying(
            connect,
            paramiko.ssh_exception.SSHException,
            retries=retries, timeout=timeout
        )
        transport.auth_publickey('root', pubkey)
        self._transport = transport

    def invalidate(self):
        # gracefully terminate transport channel
        if self._transport:
            self.m.log.debug('Closing SSH session')
            self._transport.close()
        self._transport = None
        atexit.unregister(self.invalidate)

    def _stream_out_and_err(self, channel, quiet=False):
        m_log = self.m.log.info if not quiet else self.m.log.debug
        sel = selectors.DefaultSelector()
        sel.register(channel, selectors.EVENT_READ)
        out, err, out_buf, err_buf = b'', b'', b'', b''
        last_out_time = time.time()
        silence_min = 0
        linebreak = ord(b'\n')
        while True:
            sel.select(timeout=10)
            activity = False
            while channel.recv_ready():
                r = channel.recv(16384)
                out += r
                out_buf += r
                if linebreak in r:
                    out_lines = out_buf.split(b'\n')
                    for out_line in out_lines[:-1]:
                        m_log(log.strip_control_sequences(out_line))
                    out_buf = out_lines[-1]
                activity = True
            while channel.recv_stderr_ready():
                r = channel.recv_stderr(16384)
                err += r
                err_buf += r
                if linebreak in r:
                    err_lines = err_buf.split(b'\n')
                    for err_line in err_lines[:-1]:
                        m_log(log.strip_control_sequences(err_line))
                    err_buf = err_lines[-1]
                activity = True
            if activity:
                last_out_time = time.time()
            else:
                if channel.exit_status_ready():
                    return out, err
                new_silence_min = int(time.time() - last_out_time) // 60
                if new_silence_min > silence_min:
                    silence_min = new_silence_min
                    self.m.log.debug(f'- no output for {silence_min} min -')

    def exec(self, *cmd, shell=False, quiet=False):
        cmd = (' '.join(["'" + a.replace("'", r"\'") + "'" for a in cmd])
               if not shell else cmd[0])
        self.connect()
        try:
            channel = self._transport.open_session()
        except EOFError:
            self.m.log.warning('EOFError on SSH exec, retrying in 2 sec...')
            time.sleep(2)
            self.connect(force_reconnect=True)
            channel = self._transport.open_session()
        log_func = self.m.log.debug if quiet else self.m.log.info
        for l in cmd.split('\n'):
            log_func(f'ssh: {l}')
        channel.exec_command(cmd)
        out, err = self._stream_out_and_err(channel)
        retval = channel.recv_exit_status()
        return fingertip.exec.ExecResult(retval, out, err)

    def upload(self, src, dst=None):  # may be unused
        import paramiko  # ... in parallel with VM spin-up
        assert os.path.isfile(src)
        dst = dst or os.path.basename(src)
        self.connect()
        sftp_client = paramiko.SFTPClient.from_transport(self._transport)
        sftp_client.put(src, dst)
        sftp_client.chmod(dst, os.stat(src).st_mode)

    def download(self, src, dst=None):
        import paramiko  # ... in parallel with VM spin-up
        dst = dst or os.path.basename(src)
        self.connect()
        sftp_client = paramiko.SFTPClient.from_transport(self._transport)
        sftp_client.stat(src)  # don't truncate dst if src does not exist
        sftp_client.get(src, dst)

    @property
    def key_file(self):
        s = os.stat(SSH._key_file)
        mode = s[stat.ST_MODE]
        owner = s[stat.ST_UID]
        # OpenSSH cares about permissions on key file only if the owner
        # matches current user
        if mode & 0o77 and owner == os.getuid():
            self.m.log.debug(f'fixing up permissions on {SSH._key_file}')
            os.chmod(SSH._key_file, mode & 0o7700)
        return SSH._key_file

    @property
    def pubkey(self):
        self.m.expiration.depend_on_a_file(SSH.pubkey_file)
        with open(SSH.pubkey_file) as f:
            return f.read().strip()
